@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2013 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2015 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -162,6 +162,7 @@ int FileCopy::Do()
       }
       if(get->Error() && get->Size()==0)
       {
+	 put->DontVerify();
 	 if(put->GetPos()>0)
 	 {
 	    put->PutEOF();
@@ -650,9 +651,9 @@ void FileCopy::LogTransfer()
       struct stat st;
       if(stat(fname,&st)!=-1) {
 	 if(st.st_size > long(ResMgr::Query("xfer:max-log-size",0))) {
-	    debug((9,"rotating xfer-log %s",fname));
+	    debug((9,"rotating xfer-log %s\n",fname));
 	    if(rename(fname,xstring::cat(fname,".old",NULL))==-1) {
-	       debug((1,"rename(%s): %s",fname,strerror(errno)));
+	       debug((1,"rename(%s): %s\n",fname,strerror(errno)));
 	    }
 	 }
       }
@@ -766,6 +767,35 @@ void FileCopyPeer::Seek(off_t offs)
    broken=false;
 }
 
+const char *FileCopyPeer::UseTempFile(const char *file)
+{
+   if(!ResMgr::QueryBool("xfer:use-temp-file",0))
+      return file;
+
+   xstring &temp=xstring::get_tmp(ResMgr::Query("xfer:temp-file-name",0));
+   if(temp.length()==0 || temp.eq("*"))
+      return file;
+
+   auto_rename=true;
+   temp_file=true;
+
+   const char *name=basename_ptr(file);
+   SetSuggestedFileName(name);
+
+   int subst_pos=temp.instr('*');
+   if(subst_pos>=0)
+      temp.set_substr(subst_pos,1,name);
+   else {
+      if(temp.last_char()=='.')
+	 temp.append(name);
+      else if(temp[0]=='.')
+	 temp.set_substr(0,0,name);
+      else
+	 temp.append('.').append(name);
+   }
+   return dir_file(dirname(file),temp);
+}
+
 FileCopyPeer::FileCopyPeer(dir_t m) : IOBuffer(m)
 {
    want_size=false;
@@ -785,6 +815,7 @@ FileCopyPeer::FileCopyPeer(dir_t m) : IOBuffer(m)
    range_limit=FILE_END;
    removing=false;
    file_removed=false;
+   temp_file=false;
    use_cache=true;
    write_allowed=true;
    done=false;
@@ -800,6 +831,22 @@ int FileCopyPeerFA::Do()
    int m=STALL;
    int res;
 
+   if(session->OpenMode()==FA::RENAME)
+   {
+      int res=session->Done();
+      if(res==FA::IN_PROGRESS)
+	 return m;
+      if(res<0) {
+	 if(temp_file)
+	    SetError(session->StrError(res));
+	 else
+	    debug((3,"rename failed: %s\n",session->StrError(res)));
+      }
+      session->Close();
+      done=true;
+      return MOVED;
+   }
+
    if(removing)
    {
       res=session->Done();
@@ -820,9 +867,18 @@ int FileCopyPeerFA::Do()
    if(verify)
    {
       if(verify->Error())
-	 SetError(verify->ErrorText());
-      if(verify->Done())
       {
+	 SetError(verify->ErrorText());
+	 m=MOVED;
+      }
+      else if(verify->Done())
+      {
+	 if(suggested_filename && auto_rename)
+	 {
+	    const char *new_name=dir_file(dirname(file),suggested_filename);
+	    session->Rename(file,new_name);
+	    return MOVED;
+	 }
 	 done=true;
 	 m=MOVED;
       }
@@ -999,7 +1055,8 @@ void FileCopyPeerFA::OpenSession()
    current->Timeout(0);	// mark it MOVED.
    if(mode==GET)
    {
-      if(size!=NO_SIZE && size!=NO_SIZE_YET && seek_pos>=size && !ascii)
+      if(size!=NO_SIZE && size!=NO_SIZE_YET && !ascii
+      && (seek_pos>size || (seek_pos==size && size>0)))
       {
       past_eof:
 	 debug((10,"copy src: seek past eof (seek_pos=%lld, size=%lld)\n",
@@ -1072,7 +1129,7 @@ void FileCopyPeerFA::WantSize()
    struct stat st;
    if(!strcmp(session->GetProto(),"file")
    && stat(dir_file(session->GetCwd(),file),&st)!=-1)
-      SetSize(st.st_size);
+      SetSize(S_ISREG(st.st_mode)?st.st_size:NO_SIZE);
    else
       super::WantSize();
 }
@@ -1282,6 +1339,8 @@ void FileCopyPeerFA::Init()
    can_seek0=true;
    if(FAmode==FA::LIST || FAmode==FA::LONG_LIST)
       Save(FileAccess::cache->SizeLimit());
+   if(mode==PUT)
+      file.set(UseTempFile(file));
 }
 
 FileCopyPeerFA::FileCopyPeerFA(FileAccess *s,const char *f,int m)
@@ -1341,12 +1400,12 @@ FileCopyPeer *FileCopyPeerFA::Clone()
 #define super FileCopyPeer
 
 FileCopyPeerFDStream::FileCopyPeerFDStream(FDStream *o,dir_t m)
-   : FileCopyPeer(m), my_stream(o?o:new FDStream(1,"<stdout>")), stream(my_stream)
+   : FileCopyPeer(m), my_stream(o?o:new FDStream(1,"<stdout>")), stream(my_stream), close_when_done(o!=0)
 {
    Init();
 }
 FileCopyPeerFDStream::FileCopyPeerFDStream(const Ref<FDStream>& o,dir_t m)
-   : FileCopyPeer(m), stream(o)
+   : FileCopyPeer(m), stream(o), close_when_done(false)
 {
    Init();
 }
@@ -1371,6 +1430,8 @@ void FileCopyPeerFDStream::Init()
       write_allowed=false;
    if(mode==PUT)
       put_ll_timer=new Timer(0,200);
+   if(mode==PUT && stream->fd==-1 && stream->can_setmtime())
+      stream->full_name.set(UseTempFile(stream->full_name));
 }
 
 void FileCopyPeerFDStream::Seek_LL()
@@ -1453,17 +1514,31 @@ int FileCopyPeerFDStream::Do()
    if(verify)
    {
       if(verify->Error())
+      {
 	 SetError(verify->ErrorText());
-      if(verify->Done())
+	 m=MOVED;
+      }
+      else if(verify->Done())
       {
 	 if(suggested_filename && stream && stream->full_name && auto_rename)
 	 {
 	    const char *new_name=dir_file(dirname(stream->full_name),suggested_filename);
 	    struct stat st;
-	    if((lstat(new_name,&st)==-1 && errno==ENOENT) || ResMgr::QueryBool("xfer:clobber",0)) {
+	    if(temp_file || (lstat(new_name,&st)==-1 && errno==ENOENT) || ResMgr::QueryBool("xfer:clobber",0)) {
 	       debug((5,"copy: renaming `%s' to `%s'\n",stream->full_name.get(),suggested_filename.get()));
-	       if(rename(stream->full_name,new_name)==-1)
-		  debug((3,"rename(%s, %s): %s\n",stream->full_name.get(),new_name,strerror(errno)));
+	       int res=rename(stream->full_name,new_name);
+	       if(res==-1 && errno==EIO) {
+		  // FUSE with HadoopFS workaround
+		  unlink(new_name);
+		  res=rename(stream->full_name,new_name);
+	       }
+	       if(res==-1) {
+		  const char *err=xstring::format("rename(%s, %s): %s\n",stream->full_name.get(),new_name,strerror(errno));
+		  if(temp_file)
+		     SetError(err);
+		  else
+		     debug((3,"%s\n",err));
+	       }
 	    }
 	 }
 	 done=true;
@@ -1485,7 +1560,9 @@ int FileCopyPeerFDStream::Do()
       {
 	 if(eof)
 	 {
-	    getfd(); // give it a chance to create empty file. (FIXME - handle tmp errors)
+	    // make sure the stream is open - it may create an empty file.
+	    if(stream && !stream->is_closed() && getfd()==-1)
+	       return m;
 	    if(!date_set && date!=NO_DATE && do_set_date)
 	    {
 	       if(date==NO_DATE_YET)
@@ -1494,7 +1571,7 @@ int FileCopyPeerFDStream::Do()
 	       date_set=true;
 	       m=MOVED;
 	    }
-	    if(stream && my_stream && !stream->Done())
+	    if(stream && close_when_done && !stream->Done())
 	       return m;
 	    if(!verify && do_verify)
 	       verify=new FileVerificator(stream);
@@ -1608,7 +1685,7 @@ int FileCopyPeerFDStream::Get_LL(int len)
       else
       {
 	 SetDate(st.st_mtime);
-	 SetSize(st.st_size);
+	 SetSize(S_ISREG(st.st_mode)?st.st_size:NO_SIZE);
 #ifndef NATIVE_CRLF
 	 if(ascii)
 	    SetSize(NO_SIZE);
@@ -1771,15 +1848,15 @@ FgData *FileCopyPeerFDStream::GetFgData(bool fg)
 void FileCopyPeerFDStream::WantSize()
 {
    struct stat st;
-   st.st_size=NO_SIZE;
+   int res=-1;
 
    if(stream->fd!=-1)
-      fstat(stream->fd,&st);
+      res=fstat(stream->fd,&st);
    else if(stream->full_name)
-      stat(stream->full_name,&st);
+      res=stat(stream->full_name,&st);
 
-   if(st.st_size!=NO_SIZE)
-      SetSize(st.st_size);
+   if(res!=-1)
+      SetSize(S_ISREG(st.st_mode)?st.st_size:NO_SIZE);
    else
       super::WantSize();
 }
