@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2012 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2015 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -37,6 +37,7 @@ class Resource;
 
 struct ResType
 {
+   static bool class_inited;
    static xmap<ResType*> *types_by_name;
 
    const char *name;
@@ -111,9 +112,7 @@ public:
    static const char *RangeValidate(xstring_c *value);
    static const char *ERegExpValidate(xstring_c *value);
    static const char *IPv4AddrValidate(xstring_c *value);
-#if INET6
    static const char *IPv6AddrValidate(xstring_c *value);
-#endif
    static const char *UNumberPairValidate(xstring_c *value);
    static const char *FileAccessible(xstring_c *value,int mode,bool want_dir=false);
    static const char *FileReadable(xstring_c *value);
@@ -169,6 +168,7 @@ public:
 	 return s;
       }
    bool is_nil() const { return s==0; }
+   bool is_empty() const { return s==0 || *s==0; }
    void ToNumberPair(int &a,int &b) const;
 };
 
