@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2013 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2015 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -263,13 +263,14 @@ int SFtp::Do()
       if(s==0)
       {
 	 // no more data, set attributes and close the file.
-	 if(entity_date!=NO_DATE)
-	 {
-	    Request_FSETSTAT *req=new Request_FSETSTAT(handle,protocol_version);
+	 Request_FSETSTAT *req=new Request_FSETSTAT(handle,protocol_version);
+	 if(entity_date!=NO_DATE) {
 	    req->attrs.mtime=entity_date;
 	    req->attrs.flags|=SSH_FILEXFER_ATTR_MODIFYTIME;
-	    SendRequest(req,Expect::IGNORE);
 	 }
+	 req->attrs.size=pos;
+	 req->attrs.flags|=SSH_FILEXFER_ATTR_SIZE;
+	 SendRequest(req,Expect::IGNORE);
 	 CloseHandle(Expect::DEFAULT);
 	 state=WAITING;
 	 m=MOVED;
@@ -608,8 +609,11 @@ void SFtp::SendRequest()
 	 SetError(NOT_SUPP);
 	 break;
       }
+      unsigned options=SSH_FXF_RENAME_NATIVE;
+      if(QueryBool("xfer:clobber",0))
+	 options|=SSH_FXF_RENAME_OVERWRITE;
       SendRequest(new Request_RENAME(WirePath(file),WirePath(file1),
-			SSH_FXF_RENAME_NATIVE,protocol_version),Expect::DEFAULT);
+			options,protocol_version),Expect::DEFAULT);
       state=WAITING;
       break;
    }
@@ -849,17 +853,10 @@ void SFtp::HandleExpect(Expect *e)
 	    handle_x.appendf("%02X",(unsigned char)handle[i]);
 	 LogNote(9,"got file handle %s (%d)",handle_x.get(),handle_len);
 	 request_pos=real_pos=pos;
-	 if(mode==RETRIEVE)
+	 if(mode==RETRIEVE) {
 	    SendRequest(new Request_FSTAT(handle,
 	       SSH_FILEXFER_ATTR_SIZE|SSH_FILEXFER_ATTR_MODIFYTIME|SSH_FILEXFER_ATTR_PERMISSIONS,
 	       protocol_version),Expect::INFO);
-	 else if(mode==STORE && pos>0)
-	 {
-	    // truncate the file at write position.
-	    Request_FSETSTAT *req=new Request_FSETSTAT(handle,protocol_version);
-	    req->attrs.size=pos;
-	    req->attrs.flags|=SSH_FILEXFER_ATTR_SIZE;
-	    SendRequest(req,Expect::IGNORE);
 	 }
       }
       else
@@ -915,11 +912,7 @@ void SFtp::HandleExpect(Expect *e)
 	 for(int i=0; i<r->GetCount(); i++)
 	 {
 	    const NameAttrs *a=r->GetNameAttrs(i);
-	    if(!file_set)
-	       file_set=new FileSet;
 	    FileInfo *info=MakeFileInfo(a);
-	    if(info)
-	       file_set->Add(info);
 	    if(mode==LIST)
 	    {
 	       file_buf->Put(a->name);
@@ -941,6 +934,12 @@ void SFtp::HandleExpect(Expect *e)
 		  file_buf->Put("\n");
 	       }
 	    }
+	    if(info)
+	    {
+	       if(!file_set)
+		  file_set=new FileSet;
+	       file_set->Add(info);
+	    }
 	 }
 	 if(r->Eof())
 	    goto eof;
@@ -1218,7 +1217,9 @@ int SFtp::Read(Buffer *buf,int size)
 	 return DO_AGAIN;
       if(size>size1)
 	 size=size1;
-      buf->MoveDataHere(file_buf,size);
+      size=buf->MoveDataHere(file_buf,size);
+      if(size<=0)
+	 return DO_AGAIN;
       pos+=size;
       real_pos+=size;
       rate_limit->BytesGot(size);
@@ -2054,8 +2055,6 @@ FileInfo *SFtp::MakeFileInfo(const NameAttrs *na)
 
    if(!name || !name[0])
       return 0;
-   if(strchr(name,'/'))
-      return 0;
    if(name[0]=='~')
       name=dir_file(".",name);
    Ref<FileInfo> fi(new FileInfo(name));
@@ -2290,6 +2289,7 @@ int SFtpListInfo::Do()
    if(result && session->OpenMode()!=FA::ARRAY_INFO)
    {
       ubuf=0;
+      result->ExcludeCompound();
       result->rewind();
       for(FileInfo *file=result->curr(); file!=0; file=result->next())
       {
