@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2014 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2015 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -53,20 +53,144 @@ public:
    int get_bit_length() const { return bit_length; }
    void set_bit_length(int b) { bit_length=b; set_length((b+7)/8); }
    void clear() { memset(buf,0,length()); }
+   void set_range(int from,int to,bool value);
 };
 
-struct TorrentPiece
+class TorrentBuild : public SMTask, public ProtoLog
+{
+   xstring_c top_path;
+   xstring name;
+   FileSet files;
+   StringSet dirs_to_scan;
+   bool done;
+   Ref<Error> error;
+
+   Ref<DirectedBuffer> translate;
+   const char *lc_to_utf8(const char *s);
+
+   friend class Torrent;
+   Ref<BeNode> info;
+   xstring pieces;
+   off_t total_length;
+   unsigned piece_length;
+
+   const char *CurrPath() const { return dirs_to_scan[0]; }
+   void NextDir() { dirs_to_scan.Remove(0); }
+   void QueueDir(const char *dir) { dirs_to_scan.Append(dir); }
+   void AddFile(const char *path,struct stat *st);
+   void Finish();
+
+public:
+   TorrentBuild(const char *top);
+   int Do();
+   bool Done() const { return done || error; }
+   bool Failed() const { return error; }
+   const char *ErrorText() const { return error->Text(); }
+   BeNode *GetFilesNode() const { return info->lookup("files"); }
+   void SetPiece(unsigned p,const xstring& sha1);
+   const xstring& GetMetadata();
+   const char *GetBaseDirectory() const { return dirname(top_path); }
+   const xstring& Status() const;
+};
+
+class TorrentPiece
 {
    unsigned sources_count;	    // how many peers have the piece
+   unsigned downloader_count;	    // how many downloaders of the piece are there
+   RefToArray<const TorrentPeer*> downloader; // which peers download the blocks
+   Ref<BitField> block_map;	    // which blocks are present.
 
-   BitField block_map;		    // which blocks are present
-   xarray<const TorrentPeer*> downloader; // which peers download the blocks
+public:
+   TorrentPiece() : sources_count(0), downloader_count(0) {}
+   ~TorrentPiece() {}
 
-   TorrentPiece(unsigned b)
-      : sources_count(0), block_map(b)
-      { downloader.allocate(b,0); }
+   unsigned get_sources_count() const { return sources_count; }
+   void add_sources_count(int diff) { sources_count+=diff; }
+   bool has_no_sources() const { return sources_count==0; }
 
-   bool has_a_downloader() const;
+   bool has_a_downloader() const { return downloader_count>0; }
+   void set_downloader(unsigned block,const TorrentPeer *o,const TorrentPeer *n,unsigned blk_count) {
+      if(!downloader) {
+	 if(o || !n)
+	    return;
+	 downloader=new const TorrentPeer*[blk_count];
+	 for(unsigned i=0; i<blk_count; i++)
+	    downloader[i]=0;
+      }
+      const TorrentPeer*& d=downloader[block];
+      if(d==o) {
+	 d=n;
+	 downloader_count+=(n!=0)-(o!=0);
+      }
+   }
+   void cleanup() {
+      if(downloader_count==0 && downloader)
+	 downloader=0;
+   }
+   const TorrentPeer *downloader_for(unsigned block) {
+      return downloader ? downloader[block] : 0;
+   }
+
+   void set_block_present(unsigned block,unsigned blk_count) {
+      if(!block_map)
+	 block_map=new BitField(blk_count);
+      block_map->set_bit(block,1);
+   }
+   void set_blocks_absent() {
+      block_map=0;
+   }
+   void free_block_map() {
+      block_map=0;
+   }
+   bool block_present(unsigned block) const {
+      return block_map && block_map->get_bit(block);
+   }
+   bool all_blocks_present(unsigned blk_count) const {
+      return block_map && block_map->has_all_set(0,blk_count);
+   }
+   bool any_blocks_present() const {
+      return block_map; // it's allocated when setting any bit
+   }
+};
+
+struct TorrentFile
+{
+   char *path;
+   off_t pos;
+   off_t length;
+   void set(const char *n,off_t p,off_t l) {
+      path=xstrdup(n);
+      pos=p;
+      length=l;
+   }
+   void unset() {
+      xfree(path); path=0;
+   }
+   bool contains_pos(off_t p) const {
+      return p>=pos && p<pos+length;
+   }
+};
+
+class TorrentFiles : public xarray<TorrentFile>
+{
+   static int pos_cmp(const TorrentFile *a, const TorrentFile *b) {
+      if(a->pos < b->pos)
+	 return -1;
+      if(a->pos > b->pos)
+	 return 1;
+      // we want zero-sized files to placed before non-zero ones.
+      if(a->length != b->length)
+	 return a->length < b->length ? -1 : 1;
+      return 0;
+   }
+public:
+   TorrentFile *file(int i) { return get_non_const()+i; }
+   TorrentFiles(const BeNode *f_node,const Torrent *t);
+   ~TorrentFiles() {
+      for(int i=0; i<length(); i++)
+	 file(i)->unset();
+   }
+   TorrentFile *FindByPosition(off_t p);
 };
 
 class TorrentListener : public SMTask, protected ProtoLog, protected Networker
@@ -99,6 +223,7 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    friend class TorrentPeer;
    friend class TorrentDispatcher;
    friend class TorrentListener;
+   friend class TorrentFiles;
    friend class DHT;
 
    bool shutting_down;
@@ -107,6 +232,10 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    bool is_private;
    bool validating;
    bool force_valid;
+   bool build_md;
+   bool stop_if_complete;
+   bool stop_if_known;
+   bool md_saved;
    unsigned validate_index;
    Ref<Error> invalid_cause;
 
@@ -146,8 +275,8 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    static const SMTaskRef<TorrentListener>& GetUDPSocket(const sockaddr_u& a) { return GetUDPSocket(a.family()); }
 
    static Torrent *FindTorrent(const xstring& info_hash) { return torrents.lookup(info_hash); }
-   static void AddTorrent(Torrent *t) { torrents.add(t->GetInfoHash(),t); }
-   static void RemoveTorrent(Torrent *t) { torrents.remove(t->GetInfoHash()); }
+   static void AddTorrent(Torrent *t);
+   static void RemoveTorrent(Torrent *t);
    static int GetTorrentsCount() { return torrents.count(); }
    static void Dispatch(const xstring& info_hash,int s,const sockaddr_u *remote_addr,IOBuffer *recv_buf);
    static void DispatchUDP(const char *buf,int len,const sockaddr_u& src);
@@ -157,20 +286,27 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    void FetchMetadataFromURL(const char *url);
    void StartMetadataDownload();
    void MetadataDownloaded();
-   void SetMetadata(const xstring& md);
+   bool SetMetadata(const xstring& md);
    void ParseMagnet(const char *p);
    const char *GetMetadataPath() const;
-   void SaveMetadata() const;
+   bool SaveMetadata() const;
    bool LoadMetadata(const char *path);
 
+   void Startup();
+
+   void SetTotalLength(off_t);
+   void StartValidating();
+
    xstring_c metainfo_url;
    SMTaskRef<FileCopy> metainfo_copy;
+   SMTaskRef<TorrentBuild> building;
    Ref<BeNode> metainfo_tree;
    BeNode *info;
    xstring metadata;
    xstring info_hash;
    const xstring *pieces;
    xstring name;
+   Ref<TorrentFiles> files;
 
    Ref<DirectedBuffer> recv_translate;
    Ref<DirectedBuffer> recv_translate_utf8;
@@ -192,7 +328,6 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    void AnnounceDHT();
    void DenounceDHT();
 
-   bool single_file;
    unsigned piece_length;
    unsigned last_piece_length;
    unsigned total_pieces;
@@ -214,11 +349,32 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    BeNode *Lookup(Ref<BeNode>& d,const char *name,BeNode::be_type_t type) { return Lookup(d->dict,name,type); }
 
    TaskRefArray<TorrentPeer> peers;
-   RefArray<TorrentPiece> piece_info;
    static int PeersCompareActivity(const SMTaskRef<TorrentPeer> *p1,const SMTaskRef<TorrentPeer> *p2);
    static int PeersCompareRecvRate(const SMTaskRef<TorrentPeer> *p1,const SMTaskRef<TorrentPeer> *p2);
    static int PeersCompareSendRate(const SMTaskRef<TorrentPeer> *p1,const SMTaskRef<TorrentPeer> *p2);
 
+   RefToArray<TorrentPiece> piece_info;
+   unsigned blocks_in_piece;
+   unsigned blocks_in_last_piece;
+   bool BlockPresent(unsigned piece,unsigned block) const {
+      return piece_info[piece].block_present(block);
+   }
+   bool AllBlocksPresent(unsigned piece) const {
+      return piece_info[piece].all_blocks_present(BlocksInPiece(piece));
+   }
+   bool AnyBlocksPresent(unsigned piece) const {
+      return piece_info[piece].any_blocks_present();
+   }
+   bool AllBlocksAbsent(unsigned piece) const {
+      return !AnyBlocksPresent(piece);
+   }
+   void SetBlocksAbsent(unsigned piece) {
+      piece_info[piece].set_blocks_absent();
+   }
+   void SetBlockPresent(unsigned piece,unsigned block) {
+      piece_info[piece].set_block_present(block,BlocksInPiece(piece));
+   }
+
    void RebuildPiecesNeeded();
    Timer pieces_needed_rebuild_timer;
    xarray<unsigned> pieces_needed;
@@ -260,6 +416,7 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    bool SeededEnough() const;
    float stop_on_ratio;
    Timer seed_timer;
+   Timer timeout_timer;
 
    Timer decline_timer;
    Timer optimistic_unchoke_timer;
@@ -290,12 +447,14 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    void PeerBytesGot(int b) { PeerBytesUsed(b,RateLimit::GET); }
 
    static void BlackListPeer(const TorrentPeer *peer,const char *timeout);
+   static bool BlackListed(const TorrentPeer *peer);
    TorrentPeer *FindPeerById(const xstring& p_id);
 
 public:
    static void ClassInit();
 
    Torrent(const char *mf,const char *cwd,const char *output_dir);
+   ~Torrent();
 
    int Do();
    int Done() const;
@@ -315,7 +474,7 @@ public:
    static void SHA1(const xstring& str,xstring& buf);
    void ValidatePiece(unsigned p);
    unsigned PieceLength(unsigned p) const { return p==total_pieces-1 ? last_piece_length : piece_length; }
-   unsigned BlocksInPiece(unsigned p) const { return (PieceLength(p)+BLOCK_SIZE-1)/BLOCK_SIZE; }
+   unsigned BlocksInPiece(unsigned p) const { return p==total_pieces-1 ? blocks_in_last_piece : blocks_in_piece; }
 
    const TaskRefArray<TorrentPeer>& GetPeers() const { return peers; }
    void AddPeer(TorrentPeer *);
@@ -344,6 +503,11 @@ public:
 
    void ForceValid() { force_valid=true; }
    bool IsValidating() const { return validating; }
+   void Share() { build_md=true; }
+   bool IsSharing() const { return build_md; }
+
+   void StopIfComplete() { stop_if_complete=true; }
+   void StopIfKnown() { stop_if_known=stop_if_complete=true; }
 
    static int GetPort();
    static int GetPortIPv4() { return listener?listener->GetPort():0; }
@@ -362,7 +526,7 @@ public:
 
    const TaskRefArray<TorrentTracker>& Trackers() { return trackers; }
    bool HasMetadata() const { return metadata!=0; }
-   void DisconnectPeers();
+   void RestartPeers();
 
    static void BootstrapDHT(const char *n) {
       StartDHT();
@@ -677,7 +841,7 @@ public:
 	    return res;
 	 }
       void ComputeLength() { Packet::ComputeLength(); length++; if(data) length+=data->ComputeLength(); length+=appendix.length(); }
-      void Pack(SMTaskRef<IOBuffer>& b) { Packet::Pack(b); b->PackUINT8(code); if(data) data->Pack(b); }
+      void Pack(SMTaskRef<IOBuffer>& b) { Packet::Pack(b); b->PackUINT8(code); if(data) data->Pack(b); b->Put(appendix); }
       void SetAppendix(const char *s,int len) { appendix.nset(s,len); length+=len; }
    };
 
@@ -709,6 +873,7 @@ private:
    unpack_status_t RecvHandshake();
    void SendExtensions();
    void Disconnect(const char *dc=0);
+   void Restart();
    int SendDataRequests(unsigned p);
    void SendDataRequests();
    void Have(unsigned p);
@@ -772,7 +937,7 @@ public:
 
 class TorrentBlackList
 {
-   xmap<Timer*> bl;
+   xmap_p<Timer> bl;
    void check_expire();
 public:
    bool Listed(const sockaddr_u &a);
@@ -785,11 +950,12 @@ class TorrentDispatcher : public SMTask, protected ProtoLog
    const sockaddr_u addr;
    SMTaskRef<IOBuffer> recv_buf;
    Timer timeout_timer;
+   xstring_c peer_name;
 public:
    TorrentDispatcher(int s,const sockaddr_u *a);
    ~TorrentDispatcher();
    int Do();
-   const char *GetLogContext() { return "torrent"; }
+   const char *GetLogContext() { return peer_name; }
 };
 
 #include "Job.h"
